{- Julien Cohen 2010-2011 -}

module TransfoCommands 

{- This file provides an interface to functions defined in TransforOperations (or other).    -}
{- In this interface, all functions have a list of strings as parameter.                     -}
{- The elements of that list have to be passed as separate arguments to interfaced function,
   and, if needed, transformed into integers.                                                -}
{- The list of strings as parameters is needed for interfacing with emacs 
   (at least in the HaRe architecture).                                                      -}


where 

import qualified TransfoOperations
import qualified RefacSimplifyCase
import TransfoUtils(CurryMode(..), TypeOfInterest(..))

renameToplevel args = TransfoOperations.rename_toplevel (args!!0, args!!1) (args!!2)
renameLocal args = TransfoOperations.rename_local (args!!0, args!!1) (args!!2)
renameToplevelUnsafe args = TransfoOperations.rename_toplevel_unsafe (args!!0, args!!1) (args!!2)

duplicateIntoComment args = TransfoOperations.duplicateIntoComment (args!!0) (args!!1)

generativeFold args =
    do 
      a<- TransfoOperations.select_app_of_fun_to_n_args (args!!0) ((read ( args!!1))::Int) (args!!2)
      TransfoOperations.generativeFold a (args!!2)

unfoldInstanceIn args = TransfoOperations.unfoldinstance_in (args!!0) (args!!1) (args!!2)

foldToplevelDefinition args = TransfoOperations.fold_toplevel_definition (args!!0) (args!!1)

removeDefCmd args = TransfoOperations.removedef  (args!!0) (args!!1)

simplifyCasePattern args =
    do
      c <- TransfoOperations.select_toplevel_case_in_decl (args!!0) (args!!1)
      RefacSimplifyCase.simp_op2 c (args!!1)

removeLocalDef args = TransfoOperations.removelocaldef (args!!0) (args!!1) (args!!2)

cleanImportsCmd args = TransfoOperations.cleanImports (args!!0)

rmFromExports args = TransfoOperations.rmFromExports (args!!0) (args!!1)

rmCommentBefore args = TransfoOperations.rmCommentBefore (args!!0) (args!!1)

caseToEq args =
    do
      a <- TransfoOperations.select_toplevel_decl (args!!0) (args!!1)
      RefacSimplifyCase.case_to_eq2 a (args!!1)

exhibitFunction args = TransfoOperations.exhibit_function (args!!0) (args!!1) (args!!2) (args!!3)

generalise args = 
    do let funname = args!!0
           constrname = args!!1
           localfunname = args!!2
           modname = args!!3
           (pos::Integer) = read (args!!4)
           newargname = args!!5
           (curryflag::CurryMode) = if (args!!6) == "Curried" then Curried else UnCurried
           (typeflag::TypeOfInterest)  = if (args!!7) == "RecType" then RecType else OtherType
       TransfoOperations.generalise funname constrname localfunname modname pos newargname curryflag typeflag 

generalise2 args = 
    do let funname = args!!0
           constrname = args!!1
           localfunname = args!!2
           modname = args!!3
           (pos::Integer) = read (args!!4)
           newargname = args!!5
           (curryflag::CurryMode) = if (args!!6) == "Curried" then Curried else UnCurried
           (typeflag::TypeOfInterest)  = if (args!!7) == "RecType" then RecType else OtherType
           (patternposition::Int) =read (args!!8)
       TransfoOperations.generalise2 funname constrname localfunname modname pos newargname curryflag typeflag patternposition


makeGlobalOfLocalIn args = TransfoOperations.make_global_of_local_in (args!!0) (args!!1) (args!!2)

makeGlobalOfLocal args =  TransfoOperations.make_global_of_local (args!!0) (args!!1)

generaliseIdent args = TransfoOperations.generaliseident (args!!0, args!!1) (args!!2) (args!!3)

newDefFunApp args = TransfoOperations.new_def_fun_app (args!!0) (read (args!!1)) (args!!2) (args!!3) 
newDefFunAppIn args = TransfoOperations.new_def_fun_app_in (args!!0) (args!!1) (read (args!!2)) (args!!3) (args!!4) 
newDefInfixAppIn args = TransfoOperations.new_def_infix_app_in (args!!0) (args!!1) (args!!2) (args!!3) 
newDefIdentIn args = TransfoOperations.new_def_ident_in (args!!0) (args!!1) (args!!2) (args!!3) 



moveDefBetweenModules args =  TransfoOperations.move_def_between_modules (args!!0) (args!!1) (args!!2)