{- Julien Cohen 2010-2011 -}
module TransfoOperations where

import PrettyPrint
import PosSyntax
import AbstractIO
import Maybe
import TypedIds
import UniqueNames hiding (srcLoc)
import PNT
import TiPNT
import List
import RefacUtils hiding (getParams) 
import PFE0 (findFile)
import MUtils (( # ))
import qualified RefacLocUtils
import System
import IO
import qualified RefacNewDef
import RefacGenDef
import RefacMoveDef
import RefacMvDefBtwMod
import qualified RefacFunDef -- for subFunctionDef : Fold Definition
import RefacRenaming
import qualified RefacUnsafeRenaming
import RefacDupDef
import RefacRmDef
import qualified RefacCleanImps
import qualified RefacGenFold
import TransfoUtils -- Part of the transformer layer
import qualified HsLexerPos -- for duplicateInotComment


-- Transform an argument of fold into a local binding

local_definition_of_fun_arg def_name arg_pos new_name modname = 
    do 
      filename<-PFE0.findFile (strToModName modname)
      (_, _, (mod::ModulePNT), tokList)<-parseSourceFile filename 
      let ((a,b),(c,d)) = coordinates_of_app mod tokList def_name arg_pos
      RefacNewDef.introNewDef [filename, new_name, show a , show b, show c, show d] 
 
     
-- transform a local declaration into a global one

make_global_of_local_in fun_name local_name modname =
    do filename<-PFE0.findFile (strToModName modname)
       (_, _, (mod::ModulePNT), tokList)<-parseSourceFile filename
       let ((l,c),(_,_)) = coordinates_of_local_definition_identifier_in mod tokList fun_name local_name 
       liftToTopLevel [filename, show l, show c]

make_global_of_local local_name modname = -- warning : only a the level directly under the top-level TODO : to complete
    do filename<-PFE0.findFile (strToModName modname)
       (_, _, (mod::ModulePNT), tokList)<-parseSourceFile filename
       let ((l,c),(_,_)) = coordinates_of_local_definition_identifier mod tokList local_name 
       liftToTopLevel [filename, show l, show c]
      

-- move a definition between modules

move_def_between_modules (name::String) srcmodname destmodname =
    do 
      fic <- PFE0.findFile (strToModName srcmodname)
      (_, _, (mod::ModulePNT), tokList)<-parseSourceFile fic
      let ((l,c),(_,_)) = coordinates_of_global_definition_identifier mod tokList name
      moveDefBtwMod[fic,destmodname,show l,show c]
      

-- Renamings 

rename_toplevel (oldname,modname) newname  =
    do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::ModulePNT), tokList)<-parseSourceFile fic
      let ((l,c),(_,_)) = coordinates_of_global_definition_identifier mod tokList oldname
      RefacRenaming.rename [fic,newname, show l, show c]

rename_toplevel_unsafe (oldname,modname) newname  =
    do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::ModulePNT), tokList)<-parseSourceFile fic
      let ((l,c),(_,_)) = coordinates_of_global_definition_identifier mod tokList oldname
      RefacUnsafeRenaming.rename_unsafe [fic,newname, show l, show c]

rename_local (oldname,modname) newname  = 
    do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::ModulePNT), tokList)<-parseSourceFile fic
      let ((l,c),(_,_)) = coordinates_of_local_definition_identifier mod tokList oldname
      RefacRenaming.rename [fic,newname, show l, show c]


-- exhibit the functions in a patterned definition


exhibit_function funname constrname newname modname =
    do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::ModulePNT), tokList)<-parseSourceFile fic
      let ((l,c),(l',c')) = coordinates_of_case_body mod tokList funname constrname
      RefacNewDef.introNewDef [fic,newname , show l, show c, show l', show c']


-- other_type means that the generalization is done on a value which type is not the type of interest
-- funname is the name of the function of interest
-- constrname is the name of the constructor which case has to be handled
-- modname is the module where the function is defined
-- pos si the position of the argument of the constructor to consider (0 for the first one)
-- newargname is the name of the new parameter after generalization
-- curryflag indicates if the arguments of contr are curried or not
-- typeflags indicates weither the element to generalize is of the type of interest or no (with a recursive call)
generalise funname constrname localfunname modname pos newargname (curryflag::CurryMode) (typeflag::TypeOfInterest) = 
    do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::ModulePNT), tokList)<-parseSourceFile fic
      let ((l,c),(l',c')) = coordinates_of_instance_of_pos_constr_arguments mod tokList funname constrname pos localfunname curryflag typeflag 0
      generaliseDef [fic,newargname, show l, show c, show l', show c']

generalise2 funname constrname localfunname modname pos newargname (curryflag::CurryMode) (typeflag::TypeOfInterest) patternposition= 
    do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::ModulePNT), tokList)<-parseSourceFile fic
      let ((l,c),(l',c')) = coordinates_of_instance_of_pos_constr_arguments mod tokList funname constrname pos localfunname curryflag typeflag patternposition
      generaliseDef [fic,newargname, show l, show c, show l', show c']


doduplicate (funname,modname) newname =
    do 
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::ModulePNT), tokList)<-parseSourceFile fic
      let ((l,c),(_,_)) = coordinates_of_global_definition_identifier mod tokList funname
      duplicateDef [fic,newname, show l, show c]
  
generaliseident (funname,modname) targetident newname =
 do 
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::ModulePNT), tokList)<-parseSourceFile fic
      case coordinates_of_instance mod tokList (Just funname) targetident of
        Just ((l,c),(l',c')) -> generaliseDef [fic,newname, show l, show c, show l', show c'] 
        Nothing -> error ("no instance of " ++ targetident ++ " found in " ++ funname ++ ".")

unfoldinstance id modname =
    do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::ModulePNT), tokList)<-parseSourceFile fic
      case coordinates_of_instance mod tokList Nothing id of
        Just ((l,c),(_,_)) -> RefacNewDef.unfoldDef [fic, show l, show c] 
        Nothing -> error ("no instance of " ++ id ++ "found in " ++ modname ++ ".")

unfoldinstance_in id_to_unfold id_decl_context modname =
    do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::ModulePNT), tokList)<-parseSourceFile fic
      case coordinates_of_instance mod tokList (Just id_decl_context) id_to_unfold of
        Just ((l,c),(_,_)) -> RefacNewDef.unfoldDef [fic, show l, show c] 
        Nothing -> error ("no instance of " ++ id_to_unfold ++ "found in " ++ modname ++ ".")

fold_toplevel_definition id modname =
    do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::ModulePNT), tokList)<-parseSourceFile fic
      case coordinates_of_toplevel_declaration mod tokList id of
        Just ((l,c),(l',c')) -> RefacFunDef.subFunctionDef [fic, show l, show c, show l', show c'] 
        Nothing -> error ("No toplevel declaration of " ++ id ++ "found in " ++ modname ++ ".")
 

new_def_fun_app funname nbargs newname modname =
  do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::ModulePNT), tokList)<-parseSourceFile fic
      let bs =  extract_all_bodies_of_module mod
      let apps = concat (map (get_apps funname) bs) 
      case apps of 
        [] -> error ("no application of " ++ funname ++ "found in " ++ modname ++ ".")
        a:r  ->  -- une seule fois devrait suffire, pas d'appel récursif, 
                 -- la nouvelle définition devrait se faire à plusieurs endroits automatiquement, si nécessaire en utilisant un folding
          do
            let Just a' = extract_app_to_n_args nbargs a
            let ((l,c),(l',c')) =RefacLocUtils.startEndLoc tokList a'
            RefacNewDef.introNewDef [fic,newname, show l, show c, show l', show c']

new_def_fun_app_in defname funname nbargs newname modname =
  do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::ModulePNT), tokList)<-parseSourceFile fic
      case find_definition defname (get_decl_list mod) of
        Nothing -> error ("no declaration of " ++ defname ++ "found in module " ++ modname ++ ".")
        Just d ->
            let bs =  extract_bodies_of_decl d
                apps = concat (map (get_apps funname) bs) 
            in case apps of 
                         [] -> error ("no application of " ++ funname ++ "found in module " ++ modname ++ " in declaration of " ++ defname ++".")
                         a:r  ->  -- une seule fois devrait suffire, pas d'appel récursif, 
                                  -- la nouvelle définition devrait se faire à plusieurs endroits automatiquement, si nécessaire en utilisant un folding
                                 do
                                   let Just a' = extract_app_to_n_args nbargs a
                                   let ((l,c),(l',c')) =RefacLocUtils.startEndLoc tokList a'
                                   RefacNewDef.introNewDef [fic,newname, show l, show c, show l', show c']

new_def_infix_app_in defname funname newname modname =
  do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::ModulePNT), tokList)<-parseSourceFile fic
      case find_definition defname (get_decl_list mod) of
        Nothing -> error ("no declaration of " ++ defname ++ "found in module " ++ modname ++ ".")
        Just d ->
            let bs =  extract_bodies_of_decl d
                apps = concat (map (get_infix_apps funname) bs) 
            in case apps of 
                         [] -> error ("no application of " ++ funname ++ "found in module " ++ modname ++ " in declaration of " ++ defname ++".")
                         a:r  ->  -- une seule fois devrait suffire, pas d'appel récursif, 
                                  -- la nouvelle définition devrait se faire à plusieurs endroits automatiquement, si nécessaire en utilisant un folding
                                 do
                                   let ((l,c),(l',c')) =RefacLocUtils.startEndLoc tokList a
                                   RefacNewDef.introNewDef [fic,newname, show l, show c, show l', show c']
      
new_def_ident_in defname ident newname modname =
  do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::ModulePNT), tokList)<-parseSourceFile fic
      case find_definition defname (get_decl_list mod) of
        Nothing -> error ("no declaration of " ++ defname ++ "found in module " ++ modname ++ ".")
        Just d ->
            case find_instance_of_ident_in_decl ident d of
              Nothing -> error ("no instance of " ++ ident ++ "found in module " ++ modname ++ " in declaration of " ++ defname ++".")
              Just e ->
                  do
                    let ((l,c),(l',c')) =RefacLocUtils.startEndLoc tokList e
                    RefacNewDef.introNewDef [fic,newname, show l, show c, show l', show c']
      

removedef name modname =   
  do
    fic <- PFE0.findFile (strToModName modname)
    (_, _, (mod::ModulePNT), tokList)<-parseSourceFile fic
    let ((l,c),(_,_)) = coordinates_of_global_definition_identifier mod tokList name
    removeDef [fic, show l, show c]

removelocaldef localname funname modname =   
  do
    fic <- PFE0.findFile (strToModName modname)
    (_, _, (mod::ModulePNT), tokList)<-parseSourceFile fic
    let ((l,c),(_,_)) = coordinates_of_local_definition_identifier_in mod tokList funname localname
    removeDef [fic, show l, show c]



duplicateIntoComment name modname = 
    do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::ModulePNT), tokList)<-parseSourceFile fic
      case find (TransfoUtils.defines name) (get_decl_list mod) of
        Nothing -> error ("No declaration for " ++ show name ++ " found in " ++ show modname ++ ".")
        Just d -> do let (s,_)   = RefacLocUtils.startEndLoc  tokList d 
                         newtoks = RefacLocUtils.insertComments (s,s) tokList ({-RefacLocUtils. ? -} prettyprint d) 
                     writeRefactoredFiles False [((fic,RefacLocUtils.modified),(newtoks,mod))]


rmCommentBefore name modname = 
    do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::ModulePNT), tokList)<-parseSourceFile fic
      case find (TransfoUtils.defines name) (get_decl_list mod) of
        Nothing -> error ("No declaration for " ++ show name ++ " found in " ++ show modname ++ ".")
        Just d -> do let (s,_)   = RefacLocUtils.startEndLoc  tokList d 
                         (a,b,c) = RefacLocUtils.splitToks (s,s) tokList 
                         a' = reverse a
                         a'' = reverse (removeComment a')
                     writeRefactoredFiles False [((fic,RefacLocUtils.modified),(a''++b++c,mod))]

    where removeComment [] = []
          removeComment ( (RefacLocUtils.Whitespace, _) : r) = removeComment r
          removeComment ( (RefacLocUtils.Comment, _)    : r) = r
          removeComment ( (RefacLocUtils.NestedComment, _)    : r) = r
          removeComment t = error ("Don't know what to do with token " ++ show t ++ " while removing comments.")


select_app_of_fun_to_n_args fun_name nbargs modname =
    do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::ModulePNT), _)<-parseSourceFile fic

      case TransfoUtils.select_apps_of_fun_to_n_args fun_name nbargs mod of
        [] -> error ("no application of " ++ fun_name ++ " to " ++ (show nbargs) ++" args found.")
        a:_ -> return a
       
select_toplevel_case_in_decl fun_name modname =
     do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::ModulePNT), _)<-parseSourceFile fic

      case TransfoUtils.select_toplevel_case_in_decl fun_name mod of
        [] -> error ("no case found in  " ++ fun_name ++".")
        a:_ -> return a

select_toplevel_decl fun_name modname =
     do
      fic <- PFE0.findFile (strToModName modname)
      (_, _, (mod::ModulePNT), _)<-parseSourceFile fic

      case find (TransfoUtils.defines fun_name) (TransfoUtils.get_decl_list mod) of
        Nothing -> error ("No declaration of  " ++ fun_name ++" found.")
        Just a -> return a

generativeFold exp modname = 
    do fic <- PFE0.findFile (strToModName modname)
       (_, _, (mod::ModulePNT), tokList)<-parseSourceFile fic
       let ((l,c),(l',c')) =RefacLocUtils.startEndLoc tokList exp
       RefacGenFold.genFold [fic, show l, show c, show l', show c']


cleanImports modname =
    do fic <- PFE0.findFile (strToModName modname)
       RefacCleanImps.cleanImports [fic]

rmFromExports funname modname =
    do fic <- PFE0.findFile (strToModName modname)
       (_, _, (mod::ModulePNT), tokList)<-parseSourceFile fic
       let ((l,c),(_,_)) = coordinates_of_ident_in_exports mod tokList funname
       RefacCleanImps.rmFromExport [fic, show l, show c]