{- Julien Cohen 2010-2011 -} 

module TransfoUtils  where

import PrettyPrint
import PosSyntax
import AbstractIO
import Maybe
import TypedIds
import UniqueNames hiding (srcLoc)
import PNT
import TiPNT
import List
import RefacUtils hiding (getParams, defines)
import PFE0 (findFile)
import MUtils (( # ))
import RefacLocUtils
import System
import IO


-- peut-être existe-t-il déjà un raccourci pour le type ci-dessous, dans le doute j'en crée un.
type ModulePNT = HsModuleI ModuleName PNT [HsDeclI PNT]

-- je mets un type monomorphe alors que la fonction est polymorphe de manière à documenter les utilisations qui sont faites dans ce module.
get_decl_list :: ModulePNT -> [HsDeclI PNT]
get_decl_list (HsModule  _ _ _ _ d) = d


-- identifiants PNT

is_ident_PNT :: String -> PNT -> Bool
is_ident_PNT x (PNT (PN (UnQual y) _) Value _) = (x==y)
is_ident_PNT x (PNT (PN (Qual _ y) _) Value _) = (x==y)
is_ident_PNT x _                                = False

get_ident_PNT (PNT (PN (UnQual y) _) Value _) = y
get_ident_PNT (PNT (PN (Qual _ y) _) Value _) = y

same_ident a b = get_ident_PNT a == get_ident_PNT b

-- fonctions sur les Match
type MatchPNT = HsMatchI PNT (HsExpI PNT) (HsPatI PNT) [HsDeclI PNT]

id_of_match (HsMatch _ x _ _ _) = x

bodies_of_match :: MatchPNT -> [HsExpI PNT]
bodies_of_match (HsMatch _ _ _ (HsBody e)  _) = [e] 
bodies_of_match (HsMatch _ _ _ (HsGuard l) _) = map (\ (_,_,e) -> e) l



----------------------------------------------------------------------
-- Fonctions sur les déclarations PatBind et FunBind                                                
----------------------------------------------------------------------                                                
                                                
-- fonctions sur les FunBind

is_fun_bind :: HsDeclI PNT -> Bool
is_fun_bind (Dec (HsFunBind _ _)) = True
is_fun_bind _ = False

is_fun_bind_with_name :: String -> HsDeclI PNT -> Bool
is_fun_bind_with_name  n (Dec (HsFunBind _ l)) = 
    is_ident_PNT n (id_of_match (head l))
is_fun_bind_with_name  n _ = False


filter_fun_binds :: String -> [HsDeclI PNT] -> [HsDeclI PNT]
filter_fun_binds n l = filter (is_fun_bind_with_name n) l


extract_pattern_list_of_decl (Dec (HsFunBind _ l )) = l -- warning

--  fonctions sur les PatBind

is_pat_bind ::  HsDeclI PNT -> Bool
is_pat_bind (Dec (HsPatBind _ _ _ _)) = True
is_pat_bind _ = False

is_pat_bind_with_name :: String -> HsDeclI PNT -> Bool
is_pat_bind_with_name  n (Dec (HsPatBind _ (Pat (HsPId (HsVar x))) _ _)) = 
     is_ident_PNT n x
is_pat_bind_with_name  n _ = False

defines (name::String) (d::HsDeclI PNT) = -- attention, defines :: PName -> HsDeclP -> Bool 	-- Defined in RefacUtils
    (is_pat_bind_with_name name d || is_fun_bind_with_name name d)

extract_ident_of_pat (Pat (HsPId (HsVar i))) = i
  
extract_ident_of_fun_decl (Dec(HsPatBind _ p _ _)) = extract_ident_of_pat p
extract_ident_of_fun_decl (Dec(HsFunBind _ l)) = id_of_match (head l)

is_fun_pat_bind x = is_fun_bind x || is_pat_bind x

-- This function returns the String corresponding to a pattern reduced to a variable identifier.
get_ident_of_pattern_variable p = get_ident_PNT (extract_ident_of_pat p)


-- extraction des bodies d'une declaration

extract_bodies_of_decl_with_name :: String -> HsDeclI PNT -> [HsExpI PNT]
extract_bodies_of_decl_with_name n (Dec (HsPatBind _ (Pat (HsPId (HsVar x))) (HsBody e) _)) =
    if is_ident_PNT n x then [e] else []
extract_bodies_of_decl_with_name n (Dec (HsFunBind _ l )) =
      if is_ident_PNT n (id_of_match (head l)) 
      then concat (map bodies_of_match  l)
      else []
extract_bodies_of_decl_with_name n _ = [] -- on ne prend pas en compte d'autres types de declarations (type par exemple)


extract_bodies_of_decl ::  HsDeclI PNT -> [HsExpI PNT]
extract_bodies_of_decl (Dec (HsPatBind _ (Pat _) (HsBody e) _)) = [e] 
extract_bodies_of_decl (Dec (HsFunBind _ l )) = concat (map bodies_of_match  l)
extract_bodies_of_decl _ = [] -- on ne prend pas en compte d'autres types de declarations (type par exemple)


--

extract_bodies_of_module :: String -> ModulePNT -> [HsExpI PNT]
extract_bodies_of_module n d = 
 concat( map (extract_bodies_of_decl_with_name n) (get_decl_list d))

extract_all_bodies_of_module m = 
  let decls = get_decl_list m in
  concat (map extract_bodies_of_decl decls)



--------------------------------------------------------------------------------------------------
-- fonctions de recherche d'une application dans une expression
--------------------------------------------------------------------------------------------------


--renvoie vrai si l'expression est une application de "funname" à n'importe quel nombre d'arguments
is_app_of funname (Exp (HsApp (Exp(HsId (HsVar x))) _)) = (is_ident_PNT funname x)
is_app_of funname (Exp (HsApp e1 _)) = is_app_of funname e1
is_app_of funname (Exp (HsParen e)) = is_app_of funname e
is_app_of _ _ = False

is_infix_app_of funname (Exp(HsInfixApp _ (HsVar i) _)) = is_ident_PNT funname i
is_infix_app_of _ _ = False

is_app (Exp (HsApp _ _)) = True
is_app _ = False

is_paren (Exp (HsParen _ )) = True
is_paren _ = False

--renvoie les sous-expressions étant des applications de funname à un ou plusieurs arguments 
-- (on ne renvoit pas les sous-expressions concernées dans des sous-expressions déjà renvoyées)
get_apps funname e = 
    if is_app_of funname e 
    then [remove_paren e] 
    else (case e of
            Exp (HsId _) -> []
            Exp (HsLit _ _) -> []
            Exp (HsInfixApp e1 _ e2) -> get_apps funname e1 ++ get_apps funname e2 -- ???
            Exp (HsApp e1 e2) -> get_apps funname e1 ++ get_apps funname e2
            Exp (HsNegApp _ e) -> get_apps funname e -- ???
            Exp (HsLambda _ e) -> get_apps funname e
            Exp (HsLet decls e) -> 
                let bodies = concat (map extract_bodies_of_decl decls)
                in let r = concat (map (get_apps funname) bodies)
                in get_apps funname e ++ r
            Exp (HsParen e) -> get_apps funname e
            Exp (HsCase mexpr alts) -> (get_apps funname mexpr) ++ concat (map (get_apps funname) (get_exprs_of_alts alts))
            Exp (HsTuple t) -> concat (map (get_apps funname) t)
            Exp (HsList l) -> concat (map (get_apps funname) l)
            Exp (HsIf e1 e2 e3) -> get_apps funname e1 ++ get_apps funname e2 ++ get_apps funname e3

          -- A TERMINER -- UTILISER UN COMBINATEUR DE PARCOURS
         )
        where remove_paren (Exp (HsParen e)) = e
              remove_paren a@(Exp e) = a

get_infix_apps funname e = 
    if is_infix_app_of funname e 
    then [e] 
    else (case e of
            Exp (HsId _) -> []
            Exp (HsLit _ _) -> []
            Exp (HsInfixApp e1 _ e2) -> get_infix_apps funname e1 ++ get_infix_apps funname e2 -- ???
            Exp (HsApp e1 e2) -> get_infix_apps funname e1 ++ get_infix_apps funname e2
            Exp (HsNegApp _ e) -> get_infix_apps funname e -- ???
            Exp (HsLambda _ e) -> get_infix_apps funname e
            Exp (HsLet decls e) -> 
                let bodies = concat (map extract_bodies_of_decl decls)
                in let r = concat (map (get_infix_apps funname) bodies)
                in get_infix_apps funname e ++ r
            Exp (HsParen e) -> get_infix_apps funname e
            Exp (HsCase mexpr alts) -> (get_infix_apps funname mexpr) ++ concat (map (get_infix_apps funname) (get_exprs_of_alts alts))
            Exp (HsTuple t) -> concat (map (get_infix_apps funname) t)
            Exp (HsList l) -> concat (map (get_infix_apps funname) l)
            Exp (HsIf e1 e2 e3) -> get_infix_apps funname e1 ++ get_infix_apps funname e2 ++ get_infix_apps funname e3

          -- A TERMINER -- UTILISER UN COMBINATEUR DE PARCOURS
         )


get_exprs_of_alt (HsAlt _ _ rhs ds) = get_exprs_of_rhs rhs ++ get_exprs_of_decls ds

get_exprs_of_alts as = concat (map get_exprs_of_alt as)

get_exprs_of_decl (Dec (HsPatBind _ _ rhs ds)) = get_exprs_of_rhs rhs ++ get_exprs_of_decls ds
get_exprs_of_decl (Dec (HsFunBind _ matchs)) = get_exprs_of_matchs matchs
                                         
get_exprs_of_decls ds = concat (map get_exprs_of_decl ds)

get_exprs_of_rhs (HsBody e) = [e]
get_exprs_of_rhs (HsGuard ((_,e1,e2):r)) = e1:e2:(get_exprs_of_rhs (HsGuard r))
get_exprs_of_rhs (HsGuard []) = []

get_exprs_of_matchs ms = concat (map get_exprs_of_match ms)

get_exprs_of_match (HsMatch _ _ _ rhs ds) = get_exprs_of_rhs rhs ++ get_exprs_of_decls ds

              

extract_arg n e = 
    let r = number_of_args e in 
    let d = r - n in
    if d >=0
    then extract_args_aux (r-n) e
    else error ("numbre d'arguments insuffisant")
        where extract_args_aux 0 (Exp (HsApp _  e2)) = e2
              extract_args_aux n (Exp (HsApp e1 _ )) = extract_args_aux (n-1) e1

              number_of_args (Exp(HsApp (Exp(HsId _)) _)) = 1
              number_of_args (Exp(HsApp (Exp(HsApp e1 e2)) _)) = 1 + number_of_args (Exp(HsApp e1 e2))




-- renvoie l'argument à la position arg_pos dans une application 
coordinates_of_app mod toklist def_name arg_pos = 
    let bodies = extract_bodies_of_module def_name mod in
    let target_bodies = concat (map (get_apps def_name) bodies) in
    if (length target_bodies) > 1 
        then error ("do_refac_fold: several instances found " ++ def_name)
        else if length target_bodies < 1
             then error ("do_refac_fold: no instance found :" ++ def_name ++ (show bodies))
             else let target_exp1 = extract_arg arg_pos (head target_bodies) in
                  startEndLoc toklist target_exp1
             

  
-- Etant donne une application a un nombre indetermine d'arguments, renvoie la sous-expression qui est une application à n arguments
extract_app_to_n_args n e = 
  let a = number_of_args e in 
  if a < n then Nothing
     else Just (remove_args (a-n) e)
  
 
number_of_args (Exp(HsApp e1 _)) =
      if is_app e1 || is_paren e1 then 1 + number_of_args e1 else 1
number_of_args (Exp (HsParen e)) = number_of_args e

remove_args 0 e = e
remove_args n (Exp (HsApp e1 _)) = remove_args (n-1) e1
remove_args n (Exp (HsParen e)) = remove_args n e


select_apps_of_fun_to_n_args :: String -> Int -> HsModuleP -> [HsExpP]
select_apps_of_fun_to_n_args fun_name nbargs mod_ast =
    let bs =  extract_all_bodies_of_module mod_ast
        apps = concat (map (get_apps fun_name) bs) 
    in   
      if apps == []
         then error ("No application of " ++ fun_name ++ " found.")
         else remove_nothings (map (extract_app_to_n_args nbargs) apps)
 

    where remove_nothings [] = []
          remove_nothings (Nothing:r) = remove_nothings r
          remove_nothings ((Just v):r) = v : (remove_nothings r)

---------------------------------------------------------------------------------
-- Arguments des constructeurs
---------------------------------------------------------------------------------

-- the following is for the extraction of definition in pattern matching (pre-1)

get_constructor_of_pattern (Pat (HsPId (HsCon (PNT (PN (UnQual n) _) _ _)))) = Just (n,[]) -- nullary constructors
get_constructor_of_pattern (Pat (HsPId (HsCon (PNT (PN (Qual _ n) _) _ _)))) = Just (n,[]) -- nullary constructors
get_constructor_of_pattern (Pat (HsPId (HsVar _))) = Nothing --variable
get_constructor_of_pattern (Pat (HsPLit _ _)) = Nothing -- literal
get_constructor_of_pattern (Pat (HsPNeg _ _)) = Nothing --negative numbers
get_constructor_of_pattern (Pat (HsPSucc _ _ _)) = Nothing --- ??
get_constructor_of_pattern (Pat (HsPInfixApp _ _ _)) = Nothing --infix constructors as ':'
get_constructor_of_pattern (Pat (HsPApp (PNT (PN (Qual _ n) _) (ConstrOf _ _) _) p)) = Just (n,p) -- NB : the left hand side of an App pattern is not a pattern.
get_constructor_of_pattern (Pat (HsPApp (PNT (PN (UnQual n) _) (ConstrOf _ _) _) p)) = Just (n,p) --
get_constructor_of_pattern (Pat (HsPTuple _ _ )) = Nothing
get_constructor_of_pattern (Pat (HsPList _ _)) = Nothing
get_constructor_of_pattern (Pat (HsPParen p)) = get_constructor_of_pattern p
get_constructor_of_pattern (Pat (HsPRec _ _)) = Nothing -- ???
get_constructor_of_pattern (Pat (HsPAsPat _ p)) = get_constructor_of_pattern p
get_constructor_of_pattern (Pat (HsPWildCard)) = Nothing
get_constructor_of_pattern (Pat (HsPIrrPat _)) = Nothing -- ???
                     
pattern_has_constructor c p = 
    case get_constructor_of_pattern p of
      Nothing -> False
      Just (n,_) -> c == n
       
pattern_on_constructor c n (HsMatch _ _ patlist _ _) = 
    pattern_has_constructor c (patlist!!n) -- we consider the pattern at position n

coordinates_of_case_body mod tokList funname constrname =
    let decls = get_decl_list mod in 
    let decl_of_target_fun = find (defines funname) decls in 
    case decl_of_target_fun of 
      Nothing -> error ("No declaration of " ++ funname ++ " found.")
      Just(Dec (HsFunBind _ matchlist)) -> 
          case find (pattern_on_constructor constrname 0) matchlist of -- the 0 here could be changed for more generality
            Nothing -> error ("No use of " ++ constrname ++ " in definition of " ++ funname ++ " found.")
            Just (HsMatch _ _ _ b _) -> startEndLoc tokList b
 
-- the following is for the extraction of definition in pattern matching (pre-2)
--------------------------------------------------------------------------------

data CurryMode = Curried | UnCurried

-- This function returns the argument of a given constructor at a given position in a given pattern
get_constr_argument_at_pos constrname pos (pat::HsPatI PNT) curryflag =
    case get_constructor_of_pattern pat of
      Nothing -> error ("Trying to get an argument on a pattern which is not convenient")
      Just (n,l) -> 
          if n == constrname 
          then case curryflag of
              Curried ->   nth l pos -- ./StrategyLib-4.0-beta/library/ContainerTheme.hs the first element is at pos 0
              UnCurried -> nth_of_tuplepat (head l) pos -- we consider there are no other arguments than the tuple
          else error ("Trying to get an argument on a pattern which is not convenient")

        where nth_of_tuplepat pat pos = 
                  case pat of
                    p@(Pat (HsPId (HsVar i))) -> if pos == 0 then p else error "The argument of this constructor is not a tuple."
                    Pat (HsPParen p) -> nth_of_tuplepat p pos
                    Pat (HsPTuple _ l) -> nth l pos -- ./StrategyLib-4.0-beta/library/ContainerTheme.hs the first element is at pos 0


---------------------------------------------------------------------------------
-- Get instances of identifiers
---------------------------------------------------------------------------------

-- returns the positions of the declaration instance of the ident local_name 
-- in the global definition of fun_name
coordinates_of_local_definition_identifier_in mod toklist fun_name local_name =
    let decls = get_decl_list mod in -- top-level declarations in mod
    let decl_of_target_fun = -- the declaration of fun_name
            case find (defines fun_name) decls  of
              Nothing -> error ("No top-level declaration of "++ fun_name ++ "found.")
              Just d -> d
    in 
     case find_definition local_name [decl_of_target_fun] of
            Nothing -> error ("Local binding of " ++ local_name ++ " not found in " ++ fun_name)
            Just d -> startEndLoc toklist (extract_ident_of_fun_decl d)


   
find_definition local_name [] = Nothing
find_definition local_name ds = 
              case find (defines local_name) ds of
                Just d -> Just d
                Nothing -> let next_level = get_all_decl_next_declist ds 
                           in find_definition local_name next_level


-- returns the positions of the declaration instance of the ident local_name 
-- in any declaration of the given module
coordinates_of_local_definition_identifier mod toklist local_name =
     let decls = get_decl_list mod in
     let fun_decls = filter (is_fun_pat_bind) decls in
     let local_ds = concat (map get_local_decls fun_decls)  in
     let target_ds = filter (\x -> defines local_name x) local_ds in
     if length target_ds < 1 
     then error ("Local binding of " ++ local_name ++ " not found.")
     else 
       (   startEndLoc toklist (extract_ident_of_fun_decl (head target_ds)) ) 
         where
             get_local_decls (Dec (HsFunBind _ l))  =  concat ( map (\ (HsMatch _ _ _ _ ds) -> ds) l )
             get_local_decls(Dec (HsPatBind _ _ _ ds)) = ds
  
-- TODO : rewrite the function coordinates_of_local_definition_identifier following the model of coordinates_of_local_definition_identifier_in
                                             

coordinates_of_global_definition_identifier mod toklist name = 
    let decls = get_decl_list mod in     
    let decl_of_target_fun = head (filter (defines name) decls) in -- un find serait mieux mais find renvoit un Maybe
    startEndLoc toklist (extract_ident_of_fun_decl decl_of_target_fun) 

coordinates_of_toplevel_declaration mod toklist name = 
    let decls = get_decl_list mod in     
    case find (defines name) decls of
      Nothing -> Nothing
      Just d -> Just(startEndLoc toklist d)




-- This function returns an instance (as an expression) of an identifier in an expression
find_instance_of_ident_in_expr :: String -> HsExpI PNT -> Maybe (HsExpI PNT)
find_instance_of_ident_in_expr n (Exp e) = 
    case e of 
      HsId (HsVar i) -> 
          if  is_ident_PNT n i
          then Just (Exp e)
          else Nothing 
      HsId (HsCon _)  -> Nothing 
      HsInfixApp e1 _ e2 ->
          case find_instance_of_ident_in_expr n e1 of
            Just e -> Just e
            Nothing -> find_instance_of_ident_in_expr n e2
      HsApp e1 e2 -> 
          case find_instance_of_ident_in_expr n e1 of
            Just e -> Just e
            Nothing -> find_instance_of_ident_in_expr n e2
      HsNegApp _ e ->
          find_instance_of_ident_in_expr n e
      HsIf e1 e2 e3 ->
          case find_instance_of_ident_in_expr n e1 of
            Just e -> Just e
            Nothing -> case find_instance_of_ident_in_expr n e2 of
                         Just e -> Just e
                         Nothing -> find_instance_of_ident_in_expr n e3
      HsParen e ->    find_instance_of_ident_in_expr n e
      HsWildCard -> Nothing
      HsLit _ _ -> Nothing
      HsLambda _ e ->  find_instance_of_ident_in_expr n e
      HsLet ds e -> 
          let res = find_instance_of_ident_in_dec_list Nothing {- we suppose we are already in the right declaration, so Nothing is passed -- TO BE CHECKED -} n ds 
              
          in case res of
               Just r -> Just r
               Nothing ->find_instance_of_ident_in_expr n e
      HsCase e1 alts -> 
          case find_instance_of_ident_in_expr n e1 of
            Just r -> Just r
            Nothing -> find_instance_of_ident_in_alts n alts

      HsDo s -> find_instance_of_ident_in_stmt n s
      HsTuple [] -> Nothing
      HsTuple (e:r) -> 
          case find_instance_of_ident_in_expr n e of
            res@(Just _) -> res
            Nothing -> find_instance_of_ident_in_expr n (Exp (HsTuple r))
            
      HsList [] -> Nothing
      HsList (e:r) -> 
          case find_instance_of_ident_in_expr n e of
            res@(Just _) -> res
            Nothing -> find_instance_of_ident_in_expr n (Exp (HsList r))

      HsLeftSection _ _ -> error "find_instance_of_ident_in_expr : To do (ls)." 
      HsRightSection _ _ -> error "find_instance_of_ident_in_expr : To do (rs)." 
      HsRecConstr _ _ _ -> error "find_instance_of_ident_in_expr : To do (cstr)." 
      HsRecUpdate _ _ _ -> error "find_instance_of_ident_in_expr : To do (updt)." 
      HsEnumFrom _ -> error "find_instance_of_ident_in_expr : To do (from)." 
      HsEnumFromTo _ _ -> error "find_instance_of_ident_in_expr : To do (from to)." 
      HsEnumFromThen _ _ -> error "find_instance_of_ident_in_expr : To do (from then)." 
      HsEnumFromThenTo _ _ _ -> error "find_instance_of_ident_in_expr : To do (from then to)." 
      HsListComp _ -> error "find_instance_of_ident_in_expr : To do (comp)." 
      HsExpTypeSig _ _ _ _ -> error "find_instance_of_ident_in_expr : To do (typesig)." 
      HsAsPat _ e -> find_instance_of_ident_in_expr n e
      HsIrrPat  e -> find_instance_of_ident_in_expr n e

find_instance_of_ident_in_stmt n (HsGenerator _ _ e s) = 
    case find_instance_of_ident_in_expr n e of 
      Nothing -> find_instance_of_ident_in_stmt n s
      r -> r

find_instance_of_ident_in_stmt n (HsQualifier e s) = 
    case find_instance_of_ident_in_expr n e of 
      Nothing -> find_instance_of_ident_in_stmt n s
      r -> r
find_instance_of_ident_in_stmt n (HsLetStmt ds s) =
    case find_instance_of_ident_in_stmt n s of 
      Nothing -> find_instance_of_ident_in_dec_list Nothing n ds
      r -> r
find_instance_of_ident_in_stmt n (HsLast e) = find_instance_of_ident_in_expr n e

--This function returns an instance of an identifier (as an expression) in a declaration 
find_instance_of_ident_in_decl :: String-> HsDeclP -> Maybe(HsExpP)

find_instance_of_ident_in_decl i (Dec(HsPatBind _ _ (HsBody e) _)) = 
    find_instance_of_ident_in_expr i e
find_instance_of_ident_in_decl i (Dec(HsFunBind _ [])) = Nothing
find_instance_of_ident_in_decl i (Dec(HsFunBind n ((HsMatch  _ _ _ (HsBody e) _):ms))) = 
    case find_instance_of_ident_in_expr i e of
      Nothing -> find_instance_of_ident_in_decl i (Dec(HsFunBind n ms))
      Just res -> Just res

     
-- The same as above but in a list of declaration
find_instance_of_ident_in_dec_list :: Maybe String -> String -> [HsDeclI PNT] -> Maybe(HsExpP)

find_instance_of_ident_in_dec_list (Just decname) i l =  -- we are interested in the declaration defining a given name. 
    case find (defines decname) l of
      Nothing -> Nothing
      Just p -> find_instance_of_ident_in_decl i p
find_instance_of_ident_in_dec_list Nothing i l = -- we are interested in all declarations of functions/patterns
      let l' = filter is_fun_pat_bind l
      in let res = map (find_instance_of_ident_in_decl i) l'
         in first_just res

first_just [] = Nothing
first_just (Nothing:rest) = first_just rest
first_just (a@(Just e):_) = a

          

find_instance_of_ident_in_alts n alts = first_just (map (find_instance_of_ident_in_alt n) alts)

find_instance_of_ident_in_alt n (HsAlt _ _ rhs ds) =
    case find_instance_of_ident_in_rhs n rhs of
      r@(Just e) -> r
      Nothing -> find_instance_of_ident_in_dec_list Nothing n ds

find_instance_of_ident_in_rhs n (HsBody e) = find_instance_of_ident_in_expr n e
find_instance_of_ident_in_rhs n (HsGuard []) = Nothing
find_instance_of_ident_in_rhs n (HsGuard ((_,e1,e2):r)) =


    case find_instance_of_ident_in_expr n e1 of
      r@(Just e) -> r
      Nothing ->case  find_instance_of_ident_in_expr n e2 of
                  r@(Just e) -> r
                  Nothing -> find_instance_of_ident_in_rhs n (HsGuard r)


-- This function returns an application of f to an instance of an identifier in an expression
find_instance_of_rec_app_ident_in_expr :: String -> String -> HsExpI PNT -> Maybe (HsExpI PNT)
find_instance_of_rec_app_ident_in_expr f n (Exp e) = 
    case e of 
      HsId (HsVar i) -> Nothing
      HsId (HsCon _)  -> Nothing 
      HsInfixApp e1 _ e2 ->
          case find_instance_of_rec_app_ident_in_expr f n e1 of
            Just e -> Just e
            Nothing -> find_instance_of_rec_app_ident_in_expr f n e2
                       
      p@(HsApp (Exp (HsId (HsVar f'))) (Exp (HsId (HsVar n')))) ->
          if is_ident_PNT f f' && is_ident_PNT n n'
          then Just (Exp p)
          else Nothing


      p@(HsApp e1 e2) -> 
          if is_app_of f (Exp p) && isJust (find_instance_of_ident_in_expr n (Exp p) ) {- it should be verified that n is not introduced in a new scope inside p -}
          then Just (Exp p)
          else
              case find_instance_of_rec_app_ident_in_expr f n e1 of
                Just e -> Just e
                Nothing -> find_instance_of_rec_app_ident_in_expr f n e2

      HsNegApp _ e ->
          find_instance_of_rec_app_ident_in_expr f n e
      HsIf e1 e2 e3 ->
          case find_instance_of_rec_app_ident_in_expr f n e1 of
            Just e -> Just e
            Nothing -> case find_instance_of_rec_app_ident_in_expr f n e2 of
                         Just e -> Just e
                         Nothing -> find_instance_of_rec_app_ident_in_expr f n e3
      HsParen e ->    find_instance_of_rec_app_ident_in_expr f n e
      HsWildCard -> Nothing
      HsLit _ _ -> Nothing
      HsLambda _ _ -> error "find_instance_of_rec_app_ident_in_expr : To do." 
      HsLet _ _ -> error "find_instance_of_rec_app_ident_in_expr : To do." 
      HsCase _ _ -> error "find_instance_of_rec_app_ident_in_expr : To do." 
      HsDo _ -> error "find_instance_of_rec_app_ident_in_expr : To do." 
      HsTuple _ -> error "find_instance_of_rec_app_ident_in_expr : To do." 
      HsList _ -> error "find_instance_of_rec_app_ident_in_expr : To do." 
      HsLeftSection _ _ -> error "find_instance_of_rec_app_ident_in_expr : To do." 
      HsRightSection _ _ -> error "find_instance_of_rec_app_ident_in_expr : To do." 
      HsRecConstr _ _ _ -> error "find_instance_of_rec_app_ident_in_expr : To do." 
      HsRecUpdate _ _ _ -> error "find_instance_of_rec_app_ident_in_expr : To do." 
      HsEnumFrom _ -> error "find_instance_of_rec_app_ident_in_expr : To do." 
      HsEnumFromTo _ _ -> error "find_instance_of_rec_app_ident_in_expr : To do." 
      HsEnumFromThen _ _ -> error "find_instance_of_rec_app_ident_in_expr : To do." 
      HsEnumFromThenTo _ _ _ -> error "find_instance_of_rec_app_ident_in_expr : To do." 
      HsListComp _ -> error "find_instance_of_rec_app_ident_in_expr : To do." 
      HsExpTypeSig _ _ _ _ -> error "find_instance_of_rec_app_ident_in_expr : To do." 
      HsAsPat _ _ -> error "find_instance_of_rec_app_ident_in_expr : To do." 
      HsIrrPat  _ -> error "find_instance_of_rec_app_ident_in_expr : To do." 


--This function returns an application of f to an instance of an identifier in a declaration 
find_instance_of_rec_app_ident_in_decl f i  (Dec(HsPatBind _ _ (HsBody e) _)) = 
    find_instance_of_rec_app_ident_in_expr f i e
find_instance_of_rec_app_ident_in_decl f i (Dec(HsFunBind n ((HsMatch  _ _ _ (HsBody e) _):ms))) = 
    
                    case (find_instance_of_rec_app_ident_in_expr f i e) of 
                      Nothing -> find_instance_of_rec_app_ident_in_decl f i (Dec(HsFunBind n (ms)))
                      Just res -> Just res
find_instance_of_rec_app_ident_in_decl f i (Dec(HsFunBind _ [])) = Nothing

    
-- The same as above but in a list of declaration, where we are interested in the declaration defining a given name.
-- patname : name of the local definition
-- funname : name of the global definition
find_instance_of_rec_app_ident_in_dec_list patname i funname l = 
    case find (defines patname) l of
      Nothing -> error (patname ++ " is not defined in the provided list of declarations.")
      Just p -> find_instance_of_rec_app_ident_in_decl funname i  p


data TypeOfInterest = RecType | OtherType



-- decl is a declaration of a function by pattern matching. 
-- We want to extract the case defined on a given constructor 
--   and retrieve the identifier as a given argument of that constructor.
-- Then we find an occurence of that identifier in the local definition of the case which is considered.
-- Then we return the coordinates of that occurence.
get_variable_at_pos_and_find_instance constrname pos localdefname funname (decl::HsDeclI PNT) curryflag typeflag patternposition =
-- first, find the appropriate case.
    case decl of Dec (HsFunBind _ matchlist) -> 
                     case find (pattern_on_constructor constrname patternposition) matchlist of
                       Nothing -> error ("No use of " ++ constrname ++ " found.")
                       Just (HsMatch _ _ patlist _ ds) -> 
                       -- patlist is the list of patterns is this case 
                       -- ds is the list of declarations bound to the case of interest for us
                        let p = patlist!!patternposition in 
                        let v = get_constr_argument_at_pos constrname pos p curryflag in
                        let i = get_ident_of_pattern_variable v in
                        case typeflag of 
                          OtherType ->
                              case find_instance_of_ident_in_dec_list (Just localdefname) i ds of
                                Nothing -> error ("No instance of " ++ i ++ " found in " ++ localdefname ++ ".")
                                Just r -> r
                          RecType ->  
                              case find_instance_of_rec_app_ident_in_dec_list localdefname i funname ds of
                                Nothing -> error ("No instance of recursive call on " ++ i ++ " found in " ++ localdefname ++ ".")
                                Just r -> r
                        

{- mod : ast -}
{- toklist : token list -}
{- funname : name of the function which definition is to be considered -}
{- constrname : name of the constructor to take into account in the pattern matching -}
{- pos : position of the argument of the considered constructor -}
{- localdefname : name of the local declaration where the instance has to be found -}
{- curryflag : are the arguments of the constructor curried -}
{- typeflag : select the identifier or an application of funname to it -}
{- patternposition : on which argument of funname the pattern matching has to be considered -}
coordinates_of_instance_of_pos_constr_arguments mod toklist (funname::String) constrname pos localdefname curryflag typeflag patternposition =
    
    let decls = get_decl_list mod in
    let decl_of_target_fun = find (defines funname) decls in 
    case decl_of_target_fun of 
      Nothing -> error ("No declaration of " ++ funname ++ " found.")
      Just d -> 
          let e = get_variable_at_pos_and_find_instance constrname pos localdefname funname d curryflag typeflag patternposition in
            startEndLoc toklist e    


coordinates_of_instance mod toklist funname ident = 
    case find_instance_of_ident_in_dec_list funname ident (get_decl_list mod) of
      Nothing -> Nothing
      Just i -> Just (startEndLoc toklist i)


----------------------------------------------------------------------------
-- explore a declaration to find a case expression
----------------------------------------------------------------------------

select_toplevel_case_in_decl fun_name mod =
    let es = extract_bodies_of_module fun_name mod 
    in concat (map select_toplevel_case_in_expr es)

        where 
          select_toplevel_case_in_expr (Exp (HsParen e)) = select_toplevel_case_in_expr e
          select_toplevel_case_in_expr a@(Exp (HsCase _ _)) = [a]
          select_toplevel_case_in_expr _ = error "todo : select_case_in_decl"
    

------------------------------------------------------------------------------
-- get all decls at the next level
------------------------------------------------------------------------------

get_all_decl_next (Exp e) = 
    case e of
      HsId _ -> []

      HsInfixApp e1 _ e2 -> get_all_decl_next e1 ++ get_all_decl_next e2
                       
      HsApp e1 e2 -> get_all_decl_next e1 ++ get_all_decl_next e2
         
      HsNegApp _ e -> get_all_decl_next e
          
      HsIf e1 e2 e3 -> get_all_decl_next e1 ++ get_all_decl_next e2 ++ get_all_decl_next e3
      HsParen e ->   get_all_decl_next e 
      HsWildCard -> []
      HsLit _ _ -> []
      HsLambda _ e -> get_all_decl_next e
      HsLet ds e -> ds ++ get_all_decl_next e
      HsCase e alts -> get_all_decl_next e ++ concat (map get_all_decl_next_alt alts)
      HsDo _ -> error "get_all_decl_next : To do (do)." 
      HsTuple l -> concat (map get_all_decl_next l)  
      HsList l ->  concat (map get_all_decl_next l)
      HsLeftSection _ _ -> error "get_all_decl_next : To do (left section)." 
      HsRightSection _ _ -> error "get_all_decl_next : To do (right section)." 
      HsRecConstr _ _ _ -> error "get_all_decl_next : To do (rec constr)." 
      HsRecUpdate _ _ _ -> error "get_all_decl_next : To do (update)." 
      HsEnumFrom _ -> error "get_all_decl_next : To do (enum from)." 
      HsEnumFromTo _ _ -> error "get_all_decl_next : To do (enum from to)." 
      HsEnumFromThen _ _ -> error "get_all_decl_next : To do (enum from then)." 
      HsEnumFromThenTo _ _ _ -> error "get_all_decl_next : To do (enum from then to)." 
      HsListComp _ -> [] -- arbitrary choice
      HsExpTypeSig _ _ _ _ -> error "get_all_decl_next : To do (type sig)." 
      HsAsPat _ _ -> error "get_all_decl_next : To do (as pat)." 
      HsIrrPat  _ -> error "get_all_decl_next : To do (irr pat)." 

get_all_decl_next_alt (HsAlt _ _ rhs ds) = get_all_decl_next_rhs rhs ++ ds

get_all_decl_next_rhs (HsBody e) = get_all_decl_next e
get_all_decl_next_rhs (HsGuard []) = []
get_all_decl_next_rhs (HsGuard ((_,e1,e2):r)) = get_all_decl_next e1 ++ get_all_decl_next e2 ++ get_all_decl_next_rhs (HsGuard r)

get_all_decl_next_dec (Dec d) = 
    case d of
      HsPatBind _ _ rhs ds -> get_all_decl_next_rhs rhs ++ ds
      HsFunBind _ matchlist -> concat (map get_all_decl_next_match matchlist)
      _ -> []

get_all_decl_next_declist ds = concat (map get_all_decl_next_dec ds)

get_all_decl_next_match (HsMatch _ _ _ rhs ds) = get_all_decl_next_rhs rhs ++ ds

get_first_decl_of n ds = find_definition n ds

{-get_first_decl_of n [] = None
get_first_decl_of n ds = 
    case (find (defines n) ds) of
      None -> let nextlevel = get_all_decl_next_declist ds in get_first_decl_of n nextlevel
      Just d -> Just d -}

-------------------------------------------------------
-- Exports 
-------------------------------------------------------

coordinates_of_ident_in_exports mod tokList funname =
    case mod of
      HsModule _ _ Nothing  _ _ -> error ("No exports in this module.")
      HsModule _ _ (Just es) _ _ -> case find (is_export_of funname) es of
                                        Nothing -> error ("No exports of " ++ funname ++ " found in this module.")
                                        Just (EntE (Var v) )->  startEndLoc tokList v 
                                        
    where is_export_of name (EntE (Var v)) = is_ident_PNT name v
          is_export_of _    _              = False